/*-
 * Copyright (C)2003..2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003..2011 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: open.c,v 1.24 2013/10/19 09:27:24 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include "mife_int.h"
#include "mife.h"

#define TRUFLAG(A) ( (A)                                                                              \
                   & ( MIFE_STRI                                                                      \
                     | MIFE_RTRY                                                                      \
                     | MIFE_BUFL                                                                      \
                     | MIFE_SLID                                                                      \
                     | MIFE_EOFL                                                                      \
                     | MIFE_FULL                                                                      \
                     | MIFE_BUFX                                                                      \
                     | BLIN_VERMASK                                                                   \
                   ) )

BLIN_flag mife_default = BLIN_VER1;

static ssize_t
mindelta(BLIN_flag flags) {
    ssize_t s;
    s = 0;
    if  (getMIFE_BUFL(flags)) s = 1 << (getMIFE_BUFL(flags) + 16);
    if  (s < MINDELTA) s = MINDELTA;
    return(s);
}

mife_descriptor *
mife_init(BLIN_flag flags) {
    mife_descriptor *mife = NULL;
    ifBLIN_FLAG = flags;

    ifBLIN_QX2("+ flags=%08X\n", flags);
    SETBDEF(flags);
    if  (!(mife = calloc(1, sizeof(mife_descriptor)))) {
        errno = ENOMEM;
        ifBLIN_QW0("no mem #1");
        goto out;
    }
    mife->flags = TRUFLAG(flags);
    mife->mindelta = mindelta(mife->flags);
    if  (!getMIFE_RTRY(mife->flags)) setMIFE_RTRY(mife->flags, NRETRY00);
out:
    ifBLIN_QX2("- %"BLIN_X" flags=%08X\n", BLIN_I(mife), mife->flags);
    return(mife);
}

int
mife_fini(mife_descriptor *mife) {
    ifBLIN_FLAG = mife ? mife->flags : mife_default;
    int ex = 0;

    if  (!mife) {
        ifBLIN_QX2("+ %"BLIN_X"\n", BLIN_I(mife));
        goto out;
    }
    ifBLIN_QX2("+ "BLIN_X" flags=%08X\n", BLIN_I(mife), mife->flags);
    if  (mife->kev) {
        free(mife->kev);
        mife->kev = NULL;
    }
    if  (mife->eol && mife->eol->trilen && mife->eol->trix) {
        free(mife->eol->trix);
        mife->eol->trix = NULL;
    }
    if  (mife->flags & MIFE_ALLC && mife->buffer) {
        free(mife->buffer);
        mife->buffer = NULL;
    }
    if  (mife->flags & MIFE_MMAP && mife->buffer) {
        if  ((ex = munmap(mife->buffer, mife->buflen))) {
            ifBLIN_QW0("#1");
        }
        mife->buffer = NULL;
    }
    if  (mife->flags & MIFE_CLOS) {
        if  ((ex = mife->a.close(mife->a.ad))) {
            ifBLIN_QW0("#2");
    }   }
    free(mife);
out:
    ifBLIN_QX2("- %d\n", ex);
    return(ex);
}

int
mife_ctl(mife_descriptor *mife, int type, ...) {
    ifBLIN_FLAG = mife ? mife->flags : mife_default;
    int ex = EX_OK;
    va_list ap;

    ifBLIN_QX2("+ %"BLIN_X" type=%d\n", BLIN_I(mife), type);
    va_start(ap, type);
    if  (!mife) goto out;
    switch (type) {
    case MIFE_CTLFLAG: {
            BLIN_flag flags = va_arg(ap, BLIN_flag);

            ifBLIN_QX2("  CTLFLAG %08X\n", flags);
            mife->flags = TRUFLAG(flags);
        }
        break;
    case MIFE_CTLFILE: {
            char *path = va_arg(ap, char *);

            ifBLIN_QX2("  CTLFILE %s\n", path);
            mife->flags = TRUFLAG(mife->flags) | MIFE_CLOS | MIFE_PIPE;
            mife->f.read = read;
            mife->f.close = close;
            if  (0 > (mife->f.fd = open(path, O_RDONLY))) {
                ifBLIN_QW0("#2");
                ex = errno;
                mife_fini(mife);
                mife = NULL;
                errno = ex;
                ex = -EX_IOERR;
                goto out;
        }   }
        break;
    case MIFE_CTLFDSC: {
            int fd = va_arg(ap, int);

            ifBLIN_QX2("  CTLFDSC %d\n", fd);
            mife->flags = TRUFLAG(mife->flags) | MIFE_PIPE;
            mife->f.read = read;
            mife->f.close = close;
            mife->f.fd = fd;
        }
        break;
    case MIFE_CTLCNST: {
            const char *src = va_arg(ap, char *);
            size_t size = va_arg(ap, size_t);
            size_t len;
            int i;

            ifBLIN_QX2("  CTLCNST %"BLIN_X"["BLIN_U"]\n", BLIN_I(src), size);
            mife->flags = TRUFLAG(mife->flags) | MIFE_INMM | MIFE_ALLC | MIFE_PIPE | MIFE_EOFL;
            mife->mindelta = 0;
            mife->a.ad = (void *)-1;
            mife->offset = 0;
            len = size;
            if  (!size && !(mife->flags & MIFE_STRI)) len = strlen(src);
            i = (mife->flags & MIFE_FULL) ? 1 : 0;
            mife->actlen = len;
            if  (!(mife->buffer = malloc(mife->actlen + i))) {
                mife_fini(mife);
                mife = NULL;
                errno = ENOMEM;
                ifBLIN_QW0("#2");
                ex = -EX_OSERR;
                goto out;
            }
            bcopy(src, mife->buffer, mife->actlen);
            if  (i) ((char *)(mife->buffer))[mife->actlen] = 0;
            mife->buflen = mife->actlen + i;
        }
        break;
    case MIFE_CTLBUFF: {
            char *src = va_arg(ap, char *);
            size_t size = va_arg(ap, size_t);
            size_t len;
            int i;

            ifBLIN_QX2("  CTLBUFF %"BLIN_X"["BLIN_U"]\n", BLIN_I(src), size);
            mife->flags = TRUFLAG(mife->flags) | MIFE_INMM | MIFE_PIPE | MIFE_EOFL;
            mife->mindelta = 0;
            mife->a.ad = (void *)-1;
            mife->offset = 0;
            len = size;
            if  (!size && !(mife->flags & MIFE_STRI)) len = strlen(src);
            i = (mife->flags & MIFE_FULL) ? 1 : 0;
            mife->buffer = src;
            mife->actlen = len;
            if  (i) ((char *)(mife->buffer))[mife->actlen] = 0;
            mife->buflen = mife->actlen + i;
        }
        break;
    case MIFE_CTLEOLI: {
            const char *eoln = va_arg(ap, char *);

            ifBLIN_QX2("  CTLEOLI %s\n", eoln);
            if  (!(mife->eol = malloc(sizeof(mife_eol)))) {
                ifBLIN_QW0("no mem MIFE_CTLEOLI");
                errno = ENOMEM;
                goto out;
            }
            bcopy(&mife_eol0, mife->eol, sizeof(mife_eol));
            if  (mife->eol && ((ex = mife->eol->init(mife, eoln)) < 0)) {
                goto out;
        }   }
        break;
    case MIFE_CTLITIM: {
            const struct timespec *timeout = va_arg(ap, struct timespec *);

            ifBLIN_QX2("  CTLITIM %d.%09d\n", timeout->tv_sec, timeout->tv_nsec);
            mife->itime.tv_sec = timeout->tv_sec;
            mife->itime.tv_nsec = timeout->tv_nsec;
        }
        break;
    case MIFE_CTLOTIM: {
            const struct timespec *timeout = va_arg(ap, struct timespec *);

            ifBLIN_QX2("  CTLOTIM %d.%09d\n", timeout->tv_sec, timeout->tv_nsec);
            mife->otime.tv_sec = timeout->tv_sec;
            mife->otime.tv_nsec = timeout->tv_nsec;
        }
        break;
    default:
        ifBLIN_QX0("Illegal command %08X", type);
    }
out:
    va_end(ap);
    ifBLIN_QX2("- %d\n", ex);
    return(ex);
}

mife_descriptor *
mife_open(BLIN_flag flags, const char *path) {
    mife_descriptor *mife = NULL;
    ifBLIN_FLAG = flags;
    int er;

    ifBLIN_QX2("+ flags=%08X, path=%s\n", flags, path);
    if  (!(mife = mife_init(flags))) {
        ifBLIN_QW0("#1");
        goto out;
    }
    if  (mife_ctl(mife, MIFE_CTLFILE, path)) {
        ifBLIN_QW0("#2");
        er = errno;
        mife_fini(mife);
        mife = NULL;
        errno = er;
        goto out;
    }
    ifBLIN_QX2( "- flags=%08X, mindelta=%"BLIN_X", fd=%d\n"
              , mife->flags, BLIN_I(mife->mindelta), mife->f.fd
              )
    ;
out:
    return(mife);
}

mife_descriptor *
mife_opef(BLIN_flag flags, int fd) {
    mife_descriptor *mife = NULL;
    ifBLIN_FLAG = flags;
    int er;

    ifBLIN_QX2("+ flags=%08X, fd=%d\n", flags, fd);
    if  (!(mife = mife_init(flags))) {
        ifBLIN_QW0("#1");
        goto out;
    }
    if  (mife_ctl(mife, MIFE_CTLFDSC, fd)) {
        ifBLIN_QW0("#2");
        er = errno;
        mife_fini(mife);
        mife = NULL;
        errno = er;
        goto out;
    }
    ifBLIN_QX2( "- flags=%08X, mindelta=%"BLIN_X", fd=%d\n"
              , mife->flags, BLIN_I(mife->mindelta), mife->f.fd
              )
    ;
out:
    return(mife);
}

mife_descriptor *
mife_opes(BLIN_flag flags, const char *src, size_t size) {
    mife_descriptor *mife = NULL;
    ifBLIN_FLAG = flags;
    int er;

    ifBLIN_QX2("+ flags=%08X, size=%"BLIN_U"\n", mife->flags, BLIN_I(size));
    if  (!(mife = mife_init(flags))) {
        ifBLIN_QW0("#1");
        goto out;
    }
    if  (mife_ctl(mife, MIFE_CTLCNST, src, size)) {
        ifBLIN_QW0("#2");
        er = errno;
        mife_fini(mife);
        mife = NULL;
        errno = er;
        goto out;
    }
    ifBLIN_QX2( "- flags=%08X, mindelta=%"BLIN_X", buflen=%"BLIN_X", actlen=%"BLIN_X"\n"
              , mife->flags
              , BLIN_I(mife->mindelta)
              , BLIN_I(mife->buflen)
              , BLIN_I(mife->actlen)
              )
    ;
out:
    return(mife);
}

mife_descriptor *
mife_opem(BLIN_flag flags, char *src, size_t size) {
    mife_descriptor *mife = NULL;
    ifBLIN_FLAG = flags;
    int er;

    ifBLIN_QX2("+ flags=%08X, size=%"BLIN_U"\n", mife->flags, BLIN_I(size));
    if  (!(mife = mife_init(flags))) {
        ifBLIN_QW0("#1");
        goto out;
    }
    if  (mife_ctl(mife, MIFE_CTLBUFF, src, size)) {
        ifBLIN_QW0("#2");
        er = errno;
        mife_fini(mife);
        mife = NULL;
        errno = er;
        goto out;
    }
    ifBLIN_QX2( "- flags=%08X, mindelta=%"BLIN_X", buflen=%"BLIN_X", actlen=%"BLIN_X"\n"
              , mife->flags
              , BLIN_I(mife->mindelta)
              , BLIN_I(mife->buflen)
              , BLIN_I(mife->actlen)
              )
    ;
out:
    return(mife);
}
__weak_reference(mife_fini, mife_close);
